'''
Copyright 2017 Recruit Institute of Technology

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from .entity_extractor import EntityExtractor
from .parser import Parser
from .google_document import GoogleDocument
import sys

def run(query_file):
    with open(query_file, 'r') as qfile:
        query = qfile.read()
    print("Query:\n%s" % query)

    parser = Parser(query)
    if not parser.is_parsed:
        print('Syntax error: ', parser.error_msg)
        return
    
    print("Rewritten query:\n", parser.toString())
        
    with open(parser.document_name, 'r') as myfile:
        document = myfile.read()
        
    extractor = EntityExtractor(GoogleDocument(document), True)
    entities = extractor.TopEntities(query)
    print("\nResults:\n")
    for entity in entities:
        print("%s\t%f" % (entity.span.text, entity.score))


# main

if len(sys.argv) < 2:
    print("Usage: python koko.py *.koko")
    exit()

query_files = sys.argv[1:]


for query_file in query_files:
    run(query_file)

    



'''
Copyright 2017 Recruit Institute of Technology

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

# encoding=utf8
# export FLASK_APP=demo.py
# python -m flask run
#

import spacy
from entity_extractor import EntityExtractor
from flask import Flask,render_template,request,url_for

app = Flask(__name__)

print("Loading SpaCy English models...")
nlp = spacy.load('en')
print("Done loading models")

@app.route('/')
def root():
    document = open('cafe.txt').read()
    query_str = """extract "Ents" x from "input.txt" if
                   (x "serves coffee" {0.5}) or
                   ("introducing" x {0.1}) or
                   (str(x) contains "coffee shop" {1}) or
                   (str(x) contains "cafe" {1})
               with threshold 0.8
            """

    return render_template('query.html',
                           document=document,
                           query=query_str)

@app.route('/query', methods=['POST', 'GET'])
def query():
    if request.method == 'POST':
        document = request.form['document']
        query_str = request.form['query']
        extractor = EntityExtractor(nlp(document), True)
        entities = extractor.TopEntities(query_str)
        return render_template('query.html',
                               document=document,
                               query=query_str,
                               entities=entities,
                               extractor=extractor)
    else:
        root()

with app.test_request_context():        
    url_for('static', filename='style.css')



'''
Copyright 2017 Recruit Institute of Technology

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import sys
import json
import hashlib
import httplib2
import os
from googleapiclient import discovery
from oauth2client.client import GoogleCredentials
from .segmenter import SegmentedDocument, Segment, SegmentedSpan

os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = './google_APIconfig.json'
        
class GoogleDocument(SegmentedDocument):

    def __init__(self, text):
        # super().__init__(text)
        self.text = text
        self.chunks = self.split_doc(self.text)
        self.chunk_syntax_responses = []
        # print(self.text)
        self.populate_tokens()
        self.populate_offset_to_pos_maps()
        self.populate_sents()
        self.populate_ents()
        self.is_parsed = True

    def populate_tokens(self):
        #print("populate_tokens")
        self.segments = []
        chunk_offset = 0
        i = 0
        for chunk in self.chunks:
            #print("chunk", i, ":", len(chunk), "chars")
            response = self.analyze_chunk_syntax(chunk)
            self.chunk_syntax_responses.append(response)
            self.segments += self.get_segments(response['tokens'], chunk_offset)
            chunk_offset += len(chunk)
            i += 1
        #for s in self.segments:
        #    print("Segment: [%d, %d) = %s" % (s.start, s.end, s.text))

    def populate_offset_to_pos_maps(self):
        pos = -1
        self.beginOffset2pos = {}
        self.endOffset2pos = {}
        for segment in self.segments:
            pos += 1
            start = segment.start
            end = segment.end
            self.beginOffset2pos[start] = pos
            self.endOffset2pos[end] = pos + 1
        
    def populate_sents(self):
        #print("populate_sents")
        self.sents = []
        chunk_offset = 0
        i = 0
        for chunk in self.chunks:
            #print("chunk", i, ":", len(chunk), "chars")
            response = self.chunk_syntax_responses[i]
            self.sents += self.get_resolved_spans(response['sentences'], chunk_offset)
            chunk_offset += len(chunk)
            i += 1
        
    def populate_ents(self):
        #print("populate_ents")
        self.ents = []
        self.ents_etype = []
        chunk_offset = 0
        i = 0
        for chunk in self.chunks:
            #print("chunk", i, ":", len(chunk), "chars")
            response = self.analyze_chunk_entities(chunk)
            for entity in response['entities']:
                self.ents += self.get_resolved_spans(entity['mentions'], chunk_offset, entity['type'].lower())
            chunk_offset += len(chunk)
            i += 1
            
        self.noun_chunks = self.ents

    def get_segments(self, objs, chunk_offset=0):
        segments = []
        for obj in objs:
            span = obj['text']
            beginOffset = chunk_offset + span['beginOffset']
            text = span['content']
            endOffset = beginOffset + len(text)
            prevSegmentEndOffset = -1 if not segments else segments[-1].end
            segments.append(Segment(self.text, 0, beginOffset, \
                                         beginOffset, endOffset, beginOffset==prevSegmentEndOffset))
        return segments
            
    def get_resolved_spans(self, objs, chunk_offset=0, syntax_type=''):
        spans = []
        for obj in objs:
            span = obj['text']
            beginOffset = chunk_offset + span['beginOffset']
            text = span['content']
            endOffset = beginOffset + len(text)
            resolved_span = self.resolve_span(beginOffset, endOffset, syntax_type)
            if resolved_span:
                spans.append(resolved_span)
        return spans
    
    def resolve_span(self, beginOffset, endOffset, syntax_type):
        if not beginOffset in self.beginOffset2pos:
            print('Misaligned begin offset: %d' % beginOffset)
            return None
        start = self.beginOffset2pos[beginOffset]
        if not endOffset in self.endOffset2pos:
            print('Misaligned end offset: %d' % endOffset)
            return None
        end = self.endOffset2pos[endOffset]
        #print("Span: [%d, %d) = %s" % (start, end, self[start:end].text))
        return SegmentedSpan(self, start, end, syntax_type)
        
    def get_service(self):
        credentials = GoogleCredentials.get_application_default()
        scoped_credentials = credentials.create_scoped(
            ['https://www.googleapis.com/auth/cloud-platform'])
        http = httplib2.Http()
        scoped_credentials.authorize(http)
        return discovery.build('language', 'v1beta1', http=http)

    def lookup_cached_response(self, method, text):
        md5 = hashlib.md5(text.encode('utf-8')).hexdigest()
        file_path = 'cache/'+method+'.'+md5
        if not os.path.exists(file_path):
            return None
        with open(file_path, 'r') as myfile:
            data = myfile.read()
        #print("Returning cached response for", method, md5)
        return json.loads(data)

    def cache_response(self, method, text, response):
        if not os.path.exists('cache'):
            os.makedirs('cache')
        md5 = hashlib.md5(text.encode('utf-8')).hexdigest()
        file_path = 'cache/'+method+'.'+md5
        with open(file_path, 'w') as myfile:
            myfile.write(json.dumps(response, indent=4, separators=(',', ': ')))
        #print("Caching response for", method, md5)
        
    def analyze_chunk_entities(self, chunk, encoding='UTF16'):
        response = self.lookup_cached_response('analyzeEntities', chunk)
        if response:
            return response
        body = {
            'document': {
                'type': 'PLAIN_TEXT',
                'content': chunk,
            },
            'encodingType': encoding,
        }
        service = self.get_service()
        request = service.documents().analyzeEntities(body=body)
        response = request.execute()
        self.cache_response('analyzeEntities', chunk, response)
        return response

    def analyze_chunk_syntax(self, chunk, encoding='UTF16'):
        response = self.lookup_cached_response('analyzeSyntax', chunk)
        if response:
            return response
        body = {
            'document': {
                'type': 'PLAIN_TEXT',
                'content': chunk,
            },
            'encodingType': encoding,
        }
        service = self.get_service()
        request = service.documents().analyzeSyntax(body=body)
        response = request.execute()
        #print(json.dumps(response, indent=4, separators=(',', ': ')))
        self.cache_response('analyzeSyntax', chunk, response)
        return response

    def split_doc(self, text, max_chunk_size=30000):
        if len(text) <= max_chunk_size:
            return [text]
        pos = text[:max_chunk_size].rfind('\n\n')
        if pos == -1:
            print("Couldn't split at paragraph, trying at end of line")
            pos = text[:max_chunk_size].rfind('\n')
        if pos == -1:
            print("Couldn't split at line, trying at end of word")
            pos = text[:max_chunk_size].rfind(' ')
        if pos == -1:
            pos = len(text)
            print("Couldn't find a suitable split point")
        return [text[:pos+1]] + self.split_doc(text[pos+1:], max_chunk_size)
        
        



'''
Copyright 2017 Recruit Institute of Technology

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from .entity_classifier import DictionaryEntityClassifier
from .semantic_scorer import SemanticScorer
import re
import os, sys

curr_dir = os.path.dirname(__file__)
dic_path = curr_dir + '/dict/'

class Scorer:

    def __init__(self, doc, matcher, predicates, excluding_predicates=[],
                 sentence_decomposer_server_url=''):
        self.doc = doc
        self.predicates = predicates
        self.excluding_predicates = excluding_predicates
        self.num_predicates = len(predicates)
        self.entity_classifiers = dict()
        for name in os.listdir(dic_path):
            type_name = os.path.splitext(name)[0]
            self.entity_classifiers[type_name] = \
                DictionaryEntityClassifier(dic_path+name)
        self.matcher = matcher
        self.semantic_scorer = SemanticScorer(self.doc, sentence_decomposer_server_url)

    def PredicateMatchScore(self, mention, predicate):
        span = mention.span
        n = len(predicate.context)
        context = predicate.context
        window = predicate.window
        if predicate.matching == 'semantic':
            return self.semantic_scorer.ComputeSemanticScore(mention, predicate)
        elif predicate.type == 'left':
            linked_predicate_score = 1
            if predicate.next:
                linked_predicate_score = self.PredicateMatchScore(mention, predicate.next)
            return linked_predicate_score * self.ContextMatchScore(span.start,
                                                                   context, window, -1)
        elif predicate.type == 'right':
            return self.ContextMatchScore(span.end, context, window)
        elif predicate.type == 'inside':
            window = len(span) - len(predicate.context)
            return 1 if self.ContextMatchScore(span.start, context,
                                               window, case_sensitive=True) > 0 else 0
        elif predicate.type == 'substring':
            return span.text.find(predicate.pattern) >= 0
        elif predicate.type == 'match':
            match = re.fullmatch(predicate.pattern, span.text)
            return 1 if match else 0
        elif predicate.type == 'type':
            return self.TypeMatchScore(span, context)
        elif predicate.type == 'dict':
            return self.DictionaryMatchScore(span, context)
        return 0

    def IsExcluded(self, mention):
        for predicate in self.excluding_predicates:
            if self.PredicateMatchScore(mention, predicate):
                return True
        return False
    
    def ComputeMentionScore(self, mention):
        mention.scores = []
        mention.score = 0
        if self.IsExcluded(mention):
            mention.score = -1
            return
        for i in range(self.num_predicates):
            predicate = self.predicates[i]
            predicate_score = self.PredicateMatchScore(mention, predicate)
            mention.score += predicate.weight * predicate_score
            mention.scores.append(predicate_score)

    def ScoreMentions(self, mentions):
        for mention in mentions:
            self.ComputeMentionScore(mention)

    # Aggregates entity scores from mentions scores.
    # Assummes mention scores have already been populated.
    def AggregateEntityScores(self, entity):
        entity.scores = [0 for i in range(self.num_predicates)]
        for mention in entity.mentions:
            if mention.score < 0:  # excluded mention
                return
            for i in range(self.num_predicates):
                entity.scores[i] += mention.scores[i]

    def ScoreEntities(self, entities):
        for entity in entities:
            self.AggregateEntityScores(entity)
            entity.score = 0
            for i in range(self.num_predicates):
                entity.score += self.predicates[i].weight * entity.scores[i]
            if entity.score > 1:
                entity.score = 1

    # Computes the context match score by looking for matches inside a window.
    def ContextMatchScore(self, pos, context, window, direction=1, case_sensitive=False):
        for distance in range(window + 1):
            if self.matcher.MatchesNgram(pos + direction * distance, context,
                                         direction, case_sensitive):
                return 1 / (1 + distance)
        return 0

    # Computes the dictionary match score.
    def DictionaryMatchScore(self, span, type_name):
        if not type_name in self.entity_classifiers:
            return 0
        return self.entity_classifiers[type_name].is_entity(span.text)




'''
Copyright 2017 Recruit Institute of Technology

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from .segmenter import SegmentedDocument


class TestSpan(object):

    def __init__(self, span_text, start_pos):
        self.text = span_text
        self.start_pos = start_pos


class TestDocument(SegmentedDocument):

    def __init__(self, text):
        super().__init__(text)
        # Poor man's entity extraction
        self.ents = self.__GetNgramsWithCapitalizedWords()
        # Fake noun phrase extraction
        self.noun_chunks = self.__GetNgramsWithLongWords(5)
        # Simple-minded sentence extraction.
        self.__PopulateSents()
        self.is_parsed = True

    def __GetMaximalNgrams(self, condition):
        spans = []
        prev = False
        for pos in range(len(self)):
            curr = condition(pos)
            if curr != prev:
                if curr:
                    start = pos
                else:
                    spans.append(self[start:pos])
                prev = curr
        return spans

    # Extracts maximal ngrams with tokens with at least k letters.
    def __GetNgramsWithLongWords(self, k):
        return self.__GetMaximalNgrams(lambda pos: len(self[pos].text) >= k)

    # Extracts maximal ngrams with capitalized tokens.
    def __GetNgramsWithCapitalizedWords(self):
        return self.__GetMaximalNgrams(lambda pos: self[pos].text.istitle())

    def __PopulateSents(self):
        self.sents = []
        start = 0
        for pos in range(len(self)):
            if self[pos].text == '.' or pos == len(self) - 1:
                end = pos + 1
                self.sents.append(self[start:end])
                start = end



"""
Copyright 2017 Recruit Institute of Technology

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import sys
import logging
from optparse import OptionParser

# Parse commandline arguments
op = OptionParser()
op.add_option("--query_file",
              action="store", type="string", dest="query_file",
              help="Query file name.")
op.add_option("--doc_parser",
              action="store", type="string", dest="doc_parser",
              default="koko",
              help="Text parser: koko or spacy.")
op.add_option("--output_format",
              action="store", type="string", dest="output_format",
              default="text",
              help="Output format: text or json.")
op.add_option("--log_level",
              action="store", type="string", dest="log_level",
              default="info",
              help="Logging level: info, warning, error.")
op.add_option("--target_lang",
              action="store", type="string", dest="target_lang",
              default="en",
              help="Target language: language represented by ISO 639-1 Code.")
op.add_option("--embedding_file",
              action="store", type="string", dest="embedding_file",
              default="../input/embedding.txt",
              help="The path to the embedding file.")

(opts, args) = op.parse_args()
if len(args) > 0:
    logger.error("this script takes no arguments.")

# Set up logging
logging_level_dict = { 'info': logging.INFO,
                       'warning': logging.WARNING,
                       'error': logging.ERROR }
logger = logging.getLogger()
logger.setLevel(logging.INFO)
ch = logging.StreamHandler(sys.stdout)
ch.setLevel(logging_level_dict[opts.log_level])
formatter = logging.Formatter('%(levelname)s %(asctime)s - %(message)s')
ch.setFormatter(formatter)
logger.addHandler(ch)

with open(opts.query_file, 'r') as qfile:
    query = qfile.read()

# Process the KOKO query

from query_processor import QueryProcessor
processor = QueryProcessor(opts.doc_parser, opts.target_lang)
response = processor.ProcessQuery(query, embedding_path=embedding_file)

# Print the results
if opts.output_format == 'text':
    print("\nResults:\n")
    print("%s %s" % ("{:<50}".format("Entity name"), "Entity score"))
    print("================================================================")
    for entity in response.entities:
        print("%s %f" % ("{:<50}".format(entity.name), entity.score))
else:
    import json
    import jsonpickle
    pickled = jsonpickle.encode(response, unpicklable=False)
    json_result = json.loads(pickled)
    print(json.dumps(json_result, sort_keys=False, indent=2))
    with open('json_result.txt', 'w') as ofile:
        json.dump(json_result, ofile)



'''
Copyright 2017 Recruit Institute of Technology

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

# encoding=utf8
###
# Parser for KOQL.
# Example query in KOQL:
#
#   extract NPs(x) from input.txt if
#             (x ~ "serves coffee" {0.5}) or (x ~ "has baristas" {0.5}) or
#             (x near "coffee" {3.5}) or ("check out" near x {2}) or
#             (x ", a cafe" {1}) or
#             (type(x) is "Exercises" {2}) or
#             (str(x) contains "Cafe" {1}) or (str(x) contains "Roasters" {1})
#   with threshold 0.8
###


from .segmenter import SegmentedDocument
from .query_expander import create_query_expanders, tokenize_phrase, expand_phrase

query_expanders = None

class Predicate:
    # type: what type of maching (left, right, inside, type, semantic)
    # context: what phrase to look for
    # window: how far to look for context (max number of tokens to skip)
    # weight: score multiplier
    # matching: syntactic or semantic
    # expanded_queries: a set of expanded queries through query expansion

    def __init__(self, type, context, window=0, weight = 1,
                 matching ='syntactic', expanded_queries = [], pattern=''):
        self.type = type
        self.context = context
        self.window = window
        self.weight = weight
        self.matching = matching
        self.expanded_queries = expanded_queries
        self.pattern = pattern
        self.next = None

    def toString(self, rewritten = True, weight = True):
        context = '"' + ' '.join(self.context) + '"'
        if rewritten and self.expanded_queries:
            context = ''
            for (tokens, score) in self.expanded_queries:
                if context:
                    context += ' or \n'
                context += '"%s" (%.2f)' % (tokens, score)
        
        near = ' near' if self.window else ''
        tilde = ' ~' if self.matching == 'semantic' else ''
        s = '('
        if self.type == 'left':
            s += context + near + tilde + ' x'
            if self.next:
                s += ' and ' + self.next.toString()
        elif self.type == 'right':
            s += 'x' + near + tilde + ' ' + context
        elif self.type == 'inside':
            s += 'str(x) contains ' + context
        elif self.type == 'substring':
            s += 'str(x) mentions ' + context
        elif self.type == 'match':
            s += 'str(x) matches ' + context
        if weight and self.weight != 1:
            s += ' { %.2f }' % self.weight
        s += ')'
        return s


class Parser:

    def __init__(self, query, target_lang='en', wordset=None, testing=False, embedding_path=None):
        self.query = query
        self.target_lang = target_lang
        self.wordset = wordset
        self.document_name = ''
        self.sentence_decomposer_server_url = ''
        self.predicates = []
        self.excluding_predicates = []
        self.threshold = 0
        self.lines = []
        self.curr_pos = 0
        self.error_msg = ''
        global query_expanders 
        if not query_expanders:
            query_expanders = create_query_expanders(testing, embedding_path)
        self.is_parsed = self.Parse(query)

    def toString(self):
        s = 'extract "%s" %s from "%s" if\n\t%s   \nwith threshold %.2f\n' % (
                   self.document_name,
                   self.etype,
                   self.variable_name,
                   ' or\n\t'.join([p.toString() for p in self.predicates]),
                   self.threshold )
        if self.excluding_predicates:
            s += 'excluding\n\t%s\n' % '\n\t'.join([p.toString()
                                                  for p in self.excluding_predicates])
        if self.sentence_decomposer_server_url:
            s += 'using sentence decomposition server "%s"\n' % self.sentence_decomposer_server_url
        return s

    def AtEnd(self):
        return self.curr_pos >= len(self.query_doc)

    def CurrToken(self):
        return self.query_doc[self.curr_pos]

    def Advance(self):
        if self.curr_pos < len(self.query_doc):
            self.curr_pos += 1

    def Rewind(self):
        if self.curr_pos > 0:
            self.curr_pos -= 1
            
    def Parse(self, query):
        self.lines = query.split('\n')
        self.query_doc = SegmentedDocument(query)
        try:
            self.Query()
        except AssertionError as error:
            if self.AtEnd():
                self.error_msg = "At end of query: %s" % error
            else:
                curr_token = self.CurrToken()
                self.error_msg = "Syntax error at line %d column %d: %s" % (curr_token.row,
                                                                            curr_token.col,
                                                                            error)
            return False
        return True

    # Asserts the current token has a given value, and consumes it
    def Literal(self, literal):
        assert(not self.AtEnd()), "unexpected end of query, expecting '%s'" % literal

        assert(self.Match(literal)), "unexpected token '%s', expecting '%s'" % (self.CurrToken().text,
                                                                                literal)
        self.Advance()

    # Compares the current token to a given value, but doesn't consume it
    def Match(self, literal):
        if self.AtEnd():
            return False
        return self.CurrToken().text.lower() == literal

    def Query(self):
        self.Literal("extract")
        #Determine the syntax
        if self.Match('"'):
            self.etype = self.QuotedString()[0]
            self.VariableDefinition()
        else:
            self.VariableDefinition()
            self.Literal(":")
            self.VariableType()
        self.Literal("from")
        self.DocumentName()
        self.Literal("if")
        self.Condition()
        while not self.AtEnd():
            if self.Match('with'):
                self.Threshold()
            elif self.Match('excluding'):
                self.Advance()
                self.ExcludingCondition()
            elif self.Match('using'):
                self.UsingDecompositionServer()

    def DocumentName(self):
        self.document_name = self.QuotedString()[0]

    def VariableDefinition(self):
        #assert(self.CurrToken().isidentifier()), "expecting identifier"
        self.variable_name = self.CurrToken().text
        self.Advance()
        
    def VariableType(self):
        self.etype = self.CurrToken().text
        self.Advance()

    def VariableReference(self):
        self.Literal(self.variable_name)

    def QuotedString(self):
        self.Literal('"')
        start = self.curr_pos
        while not self.AtEnd() and not self.Match('"'):
            self.Advance()
        end = self.curr_pos
        self.Literal('"')
        start_token = self.query_doc[start]
        char_start = start_token.start
        end_token = self.query_doc[end]
        char_end = end_token.start
        text = self.query[char_start:char_end]
        return tokenize_phrase(text, self.target_lang)

    def Weight(self):
        self.Literal('{')
        start = self.curr_pos
        while not self.AtEnd() and not self.Match('}'):
            self.Advance()
        end = self.curr_pos
        self.Literal('}')
        return float(self.query_doc[start:end].text)

    def Condition(self):
        self.Literal('(')
        predicate = self.Predicate()
        if self.Match('{'):
            weight = self.Weight()
            predicate.weight = weight
        self.Literal(')')
        self.predicates.append(predicate)        
        if self.Match('or'):
            self.Advance()
            self.Condition()

    def ExcludingCondition(self):
        self.Literal('(')
        predicate = self.Predicate()
        if self.Match('{'):
            weight = self.Weight()
            predicate.weight = weight
        self.Literal(')')
        self.excluding_predicates.append(predicate)        
        if self.Match('or'):
            self.Advance()
            self.ExcludingCondition()

    def Threshold(self):
        self.Literal('with')
        self.Literal('threshold')
        self.threshold = float(self.CurrToken().text)
        self.Advance()

    def UsingDecompositionServer(self):
        self.Literal('using')
        self.Literal('sentence')
        self.Literal('decomposition')
        self.sentence_decomposer_server_url = 'http://koko-demo.com:8000'
        if self.Match('server'):
            self.Advance()
            url_tokens = self.QuotedString()
            self.sentence_decomposer_server_url = ''.join(url_tokens)

    def Predicate(self):
        predicate = None
        if self.Match('"'):  # left context predicate such as: "introducing" x
            predicate = self.LeftPredicate()
        elif self.Match('str'):  # contains predicate such as: str(x) contains "cafe"
            predicate = self.InsidePredicate()
        elif self.Match('type'):   # type(x) is "Exercise"
            predicate = self.TypePredicate()
        # right context predicate such as: x ", a cafe"
        elif self.Match(self.variable_name):
            predicate = self.RightPredicate()
        else:
            assert(False), "unexpected token '%s'" % self.CurrToken().text
        global query_expanders
        if predicate.matching == 'semantic':
            predicate.expanded_queries = expand_phrase(
                query_expanders, predicate.context, self.wordset, self.target_lang)
        return predicate

    def LeftPredicate(self):
        context = self.QuotedString()
        type = 'left'
        matching='syntactic'
        window = 0
        if self.Match('~'):
            self.Advance()
            matching = 'semantic'
        elif self.Match('near'):  # proximity context predicate such as: "coffee" near x
            self.Advance()
            window = 10
        self.VariableReference()
        predicate = Predicate(type, context, window, matching=matching)
        if not self.Match(')') and not self.Match('{') :
            self.Rewind()
            predicate.next = self.Predicate()
        return predicate

    def RightPredicate(self):
        self.VariableReference()
        type = 'right'
        matching='syntactic'
        window = 0
        if self.Match('~'):
            self.Advance()
            matching = 'semantic'  # semantic context predicate such as: x ~ "had baristas"
        elif self.Match('near'):
            self.Advance()  # proximity context predicate such as: x near "coffee"
            window = 10
        context = self.QuotedString()
        return Predicate(type, context, window, matching=matching)

    def InsidePredicate(self):
        self.Literal('str')
        self.Literal('(')
        self.VariableReference()
        self.Literal(')')
        type = 'inside'
        matching='syntactic'
        if self.Match('contains'):
            self.Advance()
        elif self.Match('mentions'):
            self.Advance()
            type='substring'
        elif self.Match('matches'):
            self.Advance()
            type='match'
        elif self.Match('in'):
            self.Advance()
            self.Literal('dict')
            self.Literal('(')
            context = self.QuotedString()[0]
            self.Literal(')')
            return Predicate('dict', context)
        if self.Match('~'):
            self.Advance()
            matching = 'semantic'
        context = self.QuotedString()
        pattern = ' '.join(context)
        return Predicate(type, context, matching=matching, pattern=pattern)

    def TypePredicate(self):
        self.Literal('type')
        self.Literal('(')
        self.VariableReference()
        self.Literal(')')
        self.Literal('is')
        context = self.QuotedString()[0]
        return Predicate('type', context)



'''
Copyright 2017 Recruit Institute of Technology

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from .entity_extractor import EntityExtractor
from .parser import Parser
import spacy
from .test_document import TestDocument
import sys
from unidecode import unidecode

# main

if len(sys.argv) < 3:
    print("Usage: python koko.py query entities")
    exit()

query_file = sys.argv[1]
entities_file = sys.argv[2]


with open(query_file, 'r') as qfile:
    query = qfile.read()
# print("Query:\n%s" % query)

with open(entities_file, 'r') as efile:
    entities = set(efile.read().splitlines())

correct_entities = set()
for entity in entities:
    correct_entities.add(unidecode(entity).lower())


parser = Parser(query)
if not parser.is_parsed:
    print('Syntax error: ', parser.error_msg)
    exit()
    
#print("Rewritten query:\n", parser.toString())
        
with open(parser.document_name, 'r') as myfile:
    document = myfile.read()

print("Loading SpaCy English models...")
nlp = spacy.load('en')
print("Done loading models")
doc = nlp(document)
#doc = TestDocument(document)



extractor = EntityExtractor(doc, True)
results = extractor.TopEntities(query)

found_at_1 = set()

def matching_correct_entitity(entity):
    text = unidecode(entity.span.text).lower()
    for correct_entity in sorted(correct_entities):
        if correct_entity.startswith(text):
            return correct_entity
    return None

found_count = [0 for pred in parser.predicates]
correct_count = [0 for pred in parser.predicates]
found_entity_sets = [set() for pred in parser.predicates]
for entity in results:
    for p in range(len(parser.predicates)):
        if entity.scores[p]:
            found_count[p] += 1
            matching_entity = matching_correct_entitity(entity)
            if matching_entity:
                correct_count[p] += 1
                found_entity_sets[p].add(matching_entity)

for p in range(len(parser.predicates)):
    precision = correct_count[p] / (found_count[p] + 0.00001)
    recall = len(found_entity_sets[p]) / len(correct_entities)
    tp = correct_count[p]
    fp = found_count[p] - correct_count[p]
    print("%s pr:%6.2f re:%6.2f tp:%3d fp:%3d" % \
          ("{:<30}".format(parser.predicates[p].toString(False, False)),
           precision*100, recall*100, tp, fp))

print('Thresh\tRecall\tPrec\tF1')
for i in range(10, -1, -1):
    all_positives = 0
    true_positives = 0
    found_entities = set()
    threshold = i / 10.0
    for entity in results:
        if entity.score >= threshold:
            all_positives += 1
            matching_entity = matching_correct_entitity(entity)
            if matching_entity:
                true_positives += 1
                found_entities.add(matching_entity)
            
    precision = true_positives / (all_positives + 0.00001)
    recall = len(found_entities) / len(correct_entities)
    false_negatives = len(correct_entities) - len(found_entities)
    f1 = 2*((precision*recall)/(precision+recall + 0.00001))
    if i == 10:
        found_at_1 = found_entities
         
    print('%.3f\t%2.2f\t%2.2f\t%2.2f' % (threshold, recall*100, precision*100, f1*100))

    if i == 0:
        print('\ntp:%d fp:%d fn:%d found:%d all:%d' % (
            true_positives, all_positives - true_positives, false_negatives,
            len(found_entities), len(correct_entities) ))

"""
print('False negatives:')
for entity in correct_entities:
    if not entity in found_at_1:
        print(entity)
"""



'''
Copyright 2017 Recruit Institute of Technology

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from .parser import Parser
from .entity_extractor import EntityExtractor
from .koko_document import KokoDocument
from .google_document import GoogleDocument
import logging
import spacy

logger = logging.getLogger()

class KokoResponse:
    
    def __init__(self, query, document, entities):
        self.query = query
        self.document = document
        self.entities = entities

class QueryProcessor:
    
    def __init__(self, document_parser='koko', target_lang='en'):
        self.document_parser = document_parser
        self.target_lang = target_lang
        if self.document_parser == 'spacy':
            #logger.info("Loading SpaCy English models")
            self.nlp = spacy.load('en')
            #logger.info("Done")

    def ProcessQuery(self, query, document=None, embedding_path=None):
        query_parser = Parser(query, target_lang=self.target_lang, embedding_path=embedding_path)
        if not query_parser.is_parsed:
            logger.error("Syntax error: %s" % query_parser.error_msg)
            return None
        print("Parsed query:", query_parser.toString())
        if not document:
            with open(query_parser.document_name, 'r') as myfile:
                document = myfile.read()
        if self.document_parser == 'koko':
            doc = KokoDocument(document)
        elif self.document_parser == 'spacy':
            doc = self.nlp(document)
        elif self.document_parser == 'google':
            doc = GoogleDocument(document)
        else:
            logger.error("Unknown parser: %s" % document_parser)
            return None
        extractor = EntityExtractor(doc)
        entities = extractor.TopEntitiesForParsedQuery(query_parser)
        return KokoResponse(query, document, entities)




'''
Copyright 2017 Recruit Institute of Technology

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import unittest
from koko.matcher import Matcher
from koko.test_document import TestDocument  

class MatcherTestCase(unittest.TestCase):
    def MatchTest(self, document, pos, phrase, direction=1, case_sensitive=False):
        doc = TestDocument(document)
        matcher = Matcher(doc)
        ngram = phrase.split(' ')
        self.assertTrue(matcher.MatchesNgram(pos, ngram, direction, case_sensitive))

    def test_match(self):
        self.MatchTest('This is Cafe Benz, a full service cafe.', 2, 'Cafe Benz',
                       case_sensitive=True)

    def test_match_ignore_case(self):
        self.MatchTest('This is Cafe Benz, a full service cafe.', 2, 'cafe benz')

    def test_match_beginning(self):
        self.MatchTest('This is Cafe Benz, a full service cafe.', 0, 'this is cafe benz ,')

    def test_match_end(self):
        self.MatchTest('This is Cafe Benz, a full service cafe.', 8, 'cafe .')

    def test_reverse_match(self):
        self.MatchTest('This is Cafe Benz, a full service cafe.', 4, 'Cafe Benz', -1,
        case_sensitive=True)

    def test_reverse_match_ignore_case(self):
        self.MatchTest('This is Cafe Benz, a full service cafe.', 4, 'cafe benz', -1)

    def test_reverse_match_beginning(self):
        self.MatchTest('This is Cafe Benz, a full service cafe.', 3, 'this is cafe', -1)

    def test_reverse_match_end(self):
        self.MatchTest('This is Cafe Benz, a full service cafe.', 10, 'cafe .', -1)

if __name__ == '__main__':
    unittest.main()
    



'''
Copyright 2017 Recruit Institute of Technology

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from unidecode import unidecode

class Ngram:

    def __init__(self, doc, span):
        self.doc = doc
        self.span = span

    def tokens(self):
        return [self.doc[i].text.lower() for i in range(self.span.start, self.span.end)]

    def key(self):
        return self.span.text.lower()

    def __eq__(self, other):
        return self.key() == other.key()

    def __ne__(self, other):
        return self.key() != other.key()

    def __hash__(self):
        return hash(self.key())

    def __len__(self):
        return len(self.span)


class Matcher:

    def __init__(self, doc):
        self.doc = doc

    # Checks if the document token at position pos matches the j-th ngram token
    def TokenMatches(self, ngram, pos, j, case_sensitive=False):
        if pos < 0 or pos >= len(self.doc):
            return False
        return (self.doc[pos].text == ngram[j]) if case_sensitive else \
            (unidecode(self.doc[pos].text).lower() == unidecode(ngram[j]).lower())

    # Returns True if the document matches the ngram at position pos, False otherwise.
    # By default, direction = 1 to match at the right of pos, including the token at pos.
    # Use direction = -1 for matching at the left of pos, excluding the token
    # at pos.
    def MatchesNgram(self, pos, ngram, direction=1, case_sensitive=False):
        offset = 0 if direction > 0 else -1  # skip token at pos for reverse match
        for i in range(len(ngram)):
            j = offset + direction * i
            if not self.TokenMatches(ngram, pos + j, j, case_sensitive):
                return False
        return True

    def FindNgram(self, ngram):
        n = len(ngram)
        for pos in range(len(self.doc) - n + 1):
            if self.MatchesNgram(pos, ngram):
                return pos, pos + n
        return 0, 0



'''
Copyright 2017 Recruit Institute of Technology

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import unittest
from koko.entity_extractor import Mention, Entity
from koko.matcher import Matcher
from koko.parser import Predicate
from koko.scorer import Scorer
from koko.test_document import TestDocument


class ScorerTestCase(unittest.TestCase):

    def MentionScoreTest(self, document, phrase, predicates, expected_score=1):
        doc = TestDocument(document)
        matcher = Matcher(doc)
        scorer = Scorer(doc, matcher, predicates)
        mention = self.GetMention(doc, phrase)
        self.assertNotEqual(mention, None)
        scorer.ComputeMentionScore(mention)
        self.assertEqual(mention.score, expected_score)

    def GetMention(self, doc, phrase):
        ngram = phrase.split(' ')
        n = len(ngram)
        matcher = Matcher(doc)
        for si in range(len(doc.sents)):
                sent = doc.sents[si]
                for i in range(sent.start, sent.end - n):
                    if matcher.MatchesNgram(i, ngram):
                        return Mention(doc[i: i + len(ngram)], si)
        return None

    def test_left_context(self):
        self.MentionScoreTest('This is Cafe Benz, a full service cafe.',
                              'Cafe Benz',
                              [Predicate('left', ['this', 'is'])])

    def test_right_context(self):
        self.MentionScoreTest('This is Cafe Benz, a full service cafe.',
                              'Cafe Benz',
                              [Predicate('right', [',', 'a'])])

    def test_inside_context(self):
        self.MentionScoreTest('This is Cafe Benz, a full service cafe.',
                              'Cafe Benz',
                              [Predicate('inside', ['Cafe'])])

    def test_substring_context(self):
        self.MentionScoreTest('This is Cafe Benz, a full service cafe.',
                              'Cafe Benz',
                              [Predicate('substring', ['enz'])])

    def test_semantic_left_context(self):
        self.MentionScoreTest('This is Cafe Benz, a full service cafe.',
                              'Cafe Benz',
                              [Predicate('left', ['this', 'are'],
                                         matching='semantic',
                                         expanded_queries=[(['this', 'is'], 1.0)])])

    def text_semantic_right_context(self):
        self.MentionScoreTest('This is Cafe Benz, a full service cafe.',
                              'Cafe Benz',
                              [Predicate('right', [',', 'an'],
                                         matching='semantic',
                                         expanded_queries=[([',', 'a'], 1.0)])])

    def test_semantic_inside_context(self):
        self.MentionScoreTest('This is Cafe Benz, a full service cafe.',
                              'Cafe Benz',
                              [Predicate('inside', ['shop'],
                                         matching='semantic',
                                         expanded_queries=[(['cafe'], 1.0)])])

    def test_precedes_context(self):
        self.MentionScoreTest('This is Cafe Benz, a full service cafe.',
                              'Cafe Benz',
                              [Predicate('left', ['this'], 10)], 0.5)

    def test_follows_context(self):
        self.MentionScoreTest('This is Cafe Benz, a full service cafe.',
                              'Cafe Benz',
                              [Predicate('right', ['cafe'], 10)], 0.2)

    def test_exercise(self):
        self.MentionScoreTest('I like figure skating.',
                              'figure skating',
                              [Predicate('dict', 'Exercise')])

    def test_location(self):
        self.MentionScoreTest('I work in Mountain View.',
                              'Mountain View',
                              [Predicate('dict', 'Location')])

    def test_multiple_predicates(self):
        self.MentionScoreTest('This is Cafe Benz, a full service cafe.',
                              'Cafe Benz',
                              [
                                  Predicate('left', ['this', 'is']),
                                  Predicate('right', [',', 'a']),
                                  Predicate('inside', ['Cafe'])
                              ],
                              3)

if __name__ == '__main__':
    unittest.main()



'''
Copyright 2017 Recruit Institute of Technology

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import os, time
import operator
from collections import defaultdict
import numpy
from langdetect import detect
import progressbar
from .google_document import GoogleDocument
ROOT = os.path.abspath(os.path.dirname(__file__))

class QueryExpander(object):

    def __init__(self, lang, embedding_file_path=None, ontology_file_path=None):
        #print('Creating QueryExpander for:', lang)
        self.lang = lang
        self.__embed_vectors = None
        self.__ontology_dict = None
        self.__load_embeding_model(embedding_file_path)
        self.__load_ontology_file(ontology_file_path)

    def __load_embeding_model(self, file_path, max_vocab_size=100000):
        self.__embed_vectors = dict()
        if not file_path:
            #print('Embeddings file not provided')
            return
        if not os.path.exists(file_path):
            #print('Embeddings file not found:', file_path)
            return
        
        #print('Loading the embedding model from:', file_path)
        #bar = progressbar.ProgressBar(max_value=max_vocab_size)
        with open(file_path, "r") as embed_f:
            for line in embed_f:
                try:
                    tab = line.rstrip().split()
                    word = tab[0].lower()
                    if not word in self.__embed_vectors:
                        vec = numpy.array(tab[1:], dtype=float)
                        self.__embed_vectors[word] = vec
                except ValueError:
                    continue
                #bar.update(len(self.__embed_vectors))
                if len(self.__embed_vectors) == max_vocab_size:
                    #bar.finish()
                    return


    def __load_ontology_file(self, file_path):
        self.__ontology_dict = defaultdict(set)
        if not file_path:
            #print('Ontology file not provided')
            return
        
        if not os.path.exists(file_path):
            #print('Ontology file not found:', file_path)
            return
        
        #print('Loading ontology from:', file_path)
        with open(file_path, 'r') as ontology_f:
            for line in ontology_f:
                parent_word, child_word = line.rstrip().split('\t')
                self.__ontology_dict[parent_word].add(child_word)

    def __expand_word(self, word, topk=10, ontology_score=0.9):
        def _similarity(v1, v2):
            if v1.shape != v2.shape:
                return 0
            n2 = numpy.linalg.norm(v2)
            return float(numpy.dot(v1, v2) / n2)

        result = dict()
        result[word] = 1.0
        if self.__embed_vectors and word in self.__embed_vectors:
            word_vec = self.__embed_vectors[word]
            word_vec /= numpy.linalg.norm(word_vec)
            for cand_word, cand_vec in self.__embed_vectors.items():
                if cand_word.lower() == word.lower():
                    continue
                score = _similarity(word_vec, cand_vec)
                if cand_word in result:
                    result[cand_word] = max(score, result[cand_word])
                else:
                    result[cand_word] = score

        if word in self.__ontology_dict:
            for child_word in self.__ontology_dict[word]:
                if child_word in result:
                    result[child_word] = max(
                        ontology_score, result[child_word])
                else:
                    result[child_word] = ontology_score

        result = sorted(
            result.items(), key=operator.itemgetter(1,0), reverse=True)
        top_results = result[:min(len(result), topk)]
        return result[:min(len(result), topk)]

    def expand(self, context_tokens, wordset=None, threshold_score=0.5):
        result_list = []
        for pos, word in enumerate(context_tokens):
            word_score_pairs = self.__expand_word(word)
            next_result_list = []
            for word, score in word_score_pairs:
                if score < threshold_score:
                    continue
                if wordset and not word in wordset:
                    continue
                if pos == 0:
                    next_result_list.append(([word], score))
                    continue
                for tokens, phrase_score in result_list:
                    new_score = phrase_score * score
                    if new_score < threshold_score:
                        continue
                    next_result_list.append(
                        (tokens + [word], new_score))
            result_list = next_result_list
        result_list = sorted(
            result_list, key=operator.itemgetter(1), reverse=True)
        return result_list
       

def create_query_expanders(testing, embedding_path):
    if testing:
        query_expanders = {
            'en': QueryExpander('en')
        }
        return query_expanders
    
    #print('Loading embedding models')
    #en_embedding_file_path = ROOT + '/../embeddings/commoncrawl.840B.300d.txt'
    en_embedding_file_path = embedding_path
    ontology_file_path = ROOT + '/../coffee_ontology.txt'
    ja_embedding_file_path = ROOT + '/../embeddings/japanese_noun_verb_embedding_vectors.txt'
    query_expanders = {
        'en': QueryExpander('en', en_embedding_file_path, ontology_file_path),
        'ja': QueryExpander('ja', ja_embedding_file_path)
    }
    #print('Done loading embedding models')
    return query_expanders

def expand_phrase(query_expanders, tokens, wordset, target_lang='en'):
    phrase = " ".join(tokens)
    try:
        #lang = detect(phrase)
        lang = target_lang
    except Exception:
        lang = 'en'
    if not lang in query_expanders:
        return [(tokens, 1.0)]
    return query_expanders[lang].expand(tokens, wordset)

def tokenize_phrase(phrase, target_lang='en'):
    if len(phrase) == 0:
        return ['']
    try:
        #lang = detect(phrase)
        lang = target_lang
    except Exception:
        lang = 'en'
    if lang == 'ja':
        doc = GoogleDocument(phrase)
        tokens = [doc[i].text for i in range(len(doc))]
        #print('Japanese phrase tokens:', tokens)
    else:
        tokens = phrase.split()
    return tokens



'''
Copyright 2017 Recruit Institute of Technology

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import unittest
from koko.parser import Predicate, Parser


class ParserTestCase(unittest.TestCase):

    def ParseTest(self, query, etype, types, contexts,
                  matchings=[], weights=[], excluding_contexts=[],
                  next_types=None, next_contexts=None):
        parser = Parser(query, testing=True)
        if not parser.is_parsed:
            print("error: %s" % parser.error_msg)
        self.assertTrue(parser.is_parsed)
        n= len(parser.predicates)
        self.assertEqual(n, len(types))
        self.assertEqual(n, len(contexts))
        if weights:
            self.assertEqual(n, len(weights))
        if matchings:
            self.assertEqual(n, len(matchings))
        if next_types:
            self.assertEqual(n, len(next_types))
        if next_contexts:
            self.assertEqual(n, len(next_contexts))
        for i in range(n):
            predicate = parser.predicates[i]
            self.assertEqual(predicate.type, types[i])
            self.assertEqual(predicate.context, contexts[i])
            if weights:
                self.assertEqual(predicate.weight, weights[i])
            if matchings:
                self.assertEqual(predicate.matching, matchings[i])
        if excluding_contexts:
            self.assertEqual(len(parser.excluding_predicates), len(excluding_contexts))
            for i in range(len(parser.excluding_predicates)):
                self.assertEqual(parser.excluding_predicates[i].context,
                                 excluding_contexts[i])
        self.assertEqual(parser.etype, etype)

    def test_left_context(self):
        self.ParseTest('extract "Ents" x from "doc.txt" if ("introducing" x)'
                       'with threshold 0.8',
                       "Ents",
                       ['left'],
                       [['introducing']])

    def test_right_context(self):
        self.ParseTest('extract "Ents" x from "doc.txt" if (x "serves coffee")'
                       'with threshold 0.8',
                       "Ents",
                       ['right'],
                       [['serves', 'coffee']])

    def test_left_right_context(self):
        self.ParseTest('extract "Ents" x from "doc.txt" if ("introducing" x ", a cafe")'
                       'with threshold 0.8',
                       "Ents",
                       ['left'],
                       [['introducing']],
                       next_types=['right'],
                       next_contexts=[', a cafe'])

    def test_contains(self):
        self.ParseTest('extract "Ents" x from "doc.txt" if (str(x) contains "cafe")'
                       'with threshold 0.8',
                       "Ents",
                       ['inside'],
                       [['cafe']])

    def test_mentions(self):
        self.ParseTest('extract "Ents" x from "doc.txt" if (str(x) mentions "cafe")'
                       'with threshold 0.8',
                       "Ents",
                       ['substring'],
                       [['cafe']])

    def test_semantic_left_context(self):
        self.ParseTest('extract "Ents" x from "doc.txt" if ("introducing" ~ x)'
                       'with threshold 0.8',
                       "Ents",
                       ['left'],
                       [['introducing']],
                       ['semantic'])

    def test_semantic_right_context(self):
        self.ParseTest('extract "Ents" x from "doc.txt" if (x ~ "has baristas")'
                       'with threshold 0.8',
                       "Ents",
                       ['right'],
                       [['has', 'baristas']],
                       ['semantic'])

    def test_semantic_inside_context(self):
        self.ParseTest('extract "Ents" x from "doc.txt" if (str(x) contains ~ "cafe")'
                       'with threshold 0.8',
                       "Ents",
                       ['inside'],
                       [['cafe']],
                       ['semantic'])

    def test_dict(self):
        self.ParseTest('extract "Ents" x from "doc.txt" if (str(x) in dict("Exercise"))'
                       'with threshold 0.8',
                       "Ents",
                       ['dict'],
                       ['Exercise'])

    def test_iterator_ents(self):
        self.ParseTest('extract "Ents" x from "doc.txt" if (str(x) in dict("Exercise"))'
                       'with threshold 0.8',
                       "Ents",
                       ['dict'],
                       ['Exercise'])

    def test_iterator_nps(self):
        self.ParseTest('extract "NPs" x from "doc.txt" if (str(x) in dict("Exercise"))'
                       'with threshold 0.8',
                       "NPs",
                       ['dict'],
                       ['Exercise'])

    def test_disjunction(self):
        self.ParseTest(
            """ extract "Ents" x from "doc.txt" if
                   (str(x) contains "cafe") or
                   (x "serves coffee") or
                   ("introducing" x)
                with threshold 0.8""",
            "Ents",
            ['inside', 'right', 'left'],
            [['cafe'], ['serves', 'coffee'], ['introducing']])

    def test_weights(self):
        self.ParseTest(
            """ extract "Ents" x from "doc.txt" if
                   (str(x) contains "cafe" {1}) or
                   (x "serves coffee" {2.5}) or
                   ("introducing" x {6.5})
                with threshold 0.8""",
            "Ents",
            ['inside', 'right', 'left'],
            [['cafe'], ['serves', 'coffee'], ['introducing']],
            weights = [1, 2.5, 6.5])

    def test_excluding(self):
        self.ParseTest(
            """ extract "Ents" x from "doc.txt" if
                   (str(x) contains "cafe") or
                   (x "serves coffee") or
                   ("introducing" x)
                with threshold 0.8
                excluding (str(x) contains "the") or (x "tea") or ("food court" near x)
            """,
            "Ents",
            ['inside', 'right', 'left'],
            [['cafe'], ['serves', 'coffee'], ['introducing']],
            excluding_contexts=[['the'], ['tea'], ['food', 'court']]
        )

    def test_syntax_error(self):
        parser = Parser(
            'select Ents(x) from "doc.txt" if (x "serves coffee") with threshold 0.8')
        self.assertFalse(parser.is_parsed)
        self.assertEqual(parser.error_msg, 'Syntax error at line 1 column 0:'
                                           ' unexpected token \'select\', expecting \'extract\'')

    def test_error_non_terminated_string(self):
        parser = Parser(
            'extract "Ents" x from "doc.txt" if (str(x) contains "cafe)')
        self.assertFalse(parser.is_parsed)
        self.assertEqual(
            parser.error_msg, 'At end of query: unexpected end of query, expecting \'"\'')

'''
    def test_etype_syntax_context(self):
        self.ParseTest('extract x:Entity from "doc.txt" if ("introducing" x)'
                       'with threshold 0.8',
                       "Entity",
                       ['left'],
                       [['introducing']])
'''

if __name__ == '__main__':
    unittest.main()



'''
Copyright 2017 Recruit Institute of Technology

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from .segmenter import SegmentedDocument

class KokoDocument(SegmentedDocument):

    def __init__(self, text):
        super().__init__(text)
        # Simple-minded sentence extraction.
        self.__PopulateSents()
        # Simple heuristic for entity extraction
        self.ents = self.__GetNgramsWithCapitalizedWords()
        # Noun chunks not implemented
        self.noun_chunks = self.ents
        self.is_parsed = True

    def __GetMaximalNgrams(self, condition):
        spans = []
        prev = False
        for pos in range(len(self)):
            curr = condition(pos)
            if curr != prev:
                if curr:
                    start = pos
                else:
                    spans.append(self[start:pos])
                prev = curr
        return spans

    def __GetMaximalNgrams2(self, condition):
        spans = []
        for sent in self.sents:
            prev = False
            for pos in range(sent.start, sent.end):
                curr = condition(pos)
                if curr != prev:
                    if curr:
                        start = pos
                    else:
                        spans.append(self[start:pos])
                        print("Inside entity: ", self[start:pos].text)
                        prev = curr
            if curr:
                spans.append(self[start:pos])
                print("End entity: ", self[start:pos].text)
        return spans

    # Extracts maximal ngrams with capitalized tokens.
    def __GetNgramsWithCapitalizedWords(self):
        return self.__GetMaximalNgrams(lambda pos: self[pos].text.istitle())

    def __PopulateSents(self):
        self.sents = []
        start = 0
        for pos in range(len(self)):
            if self[pos].text == '.' or pos == len(self) - 1 or \
               self[pos + 1].row != self[pos].row:
                end = pos + 1
                self.sents.append(self[start:end])
                start = end



'''
Copyright 2017 Recruit Institute of Technology

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import json
import requests
from .koko_document import KokoDocument

class EntailedSentence:

    def __init__(self, doc, score):
        self.doc = doc
        self.score = score


class SentenceDecomposer:

    def __init__(self, server_url):
        self.server_url = server_url

    def decompose(self, doc):
        if not self.server_url:
            return [[EntailedSentence(doc, 1.0)]]
        
        r = requests.post(
            self.server_url, data={'document': doc.text.encode('utf-8')})

        if not r.text:  # Empty reply from server
            # Fall back to a single entailed sentence equal to the original text.
            return [[EntailedSentence(doc, 1.0)]]
    
        json_response = r.json()
        
        response = []
        for sentence_entry in json_response:
            entailed_sentences = []
            for entailed_sentence_entry in sentence_entry["entailment"]:
                entailed_sentence = entailed_sentence_entry[
                    "entailed_sentence"]
                score = entailed_sentence_entry["score"]
                doc = KokoDocument(entailed_sentence)
                entailed_sentences.append(EntailedSentence(doc, score))
            response.append(entailed_sentences)
        return response




'''
Copyright 2017 Recruit Institute of Technology

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

# Unit tests for EntityExtractor

import unittest
from koko.entity_extractor import EntityExtractor
# Dependency injection: use TestDocument instead of the SpaCy document to avoid loading
# the SpaCy models.
# Warning: the extracted entities and noun chunks are fake, using simple heuristics.
from koko.test_document import TestDocument  

class EntityExtractorTestCase(unittest.TestCase):
    def QueryTest(self, document, query, expected_top_entities, expected_top_scores = None):
        doc = TestDocument(document)
        extractor = EntityExtractor(doc, testing=True)
        entities = extractor.TopEntities(query)
        n = len(entities)
        self.assertGreaterEqual(n, len(expected_top_entities))
        if expected_top_scores:
            self.assertEqual(len(expected_top_entities), len(expected_top_scores))
        for i in range(len(expected_top_entities)):
            curr_entity = entities[i]
            self.assertEqual(curr_entity.span.text, expected_top_entities[i])
            if expected_top_scores:
                self.assertAlmostEqual(curr_entity.score, expected_top_scores[i])

    def test_ents_left_context(self):
        self.QueryTest('Let me introduce Cafe Benz, a full service cafe '
                       'in the middle of the car dealership.'
                       'Sit on a soft leather couch in Cafe Benz.',
                       'extract "Ents" x from "doc.txt" if ("introduce" x)'
                       'with threshold 0.8',
                       ['Cafe Benz'])

    def test_ents_right_context(self):
        self.QueryTest('This is Cafe Benz, a full service cafe.',
                       'extract "Ents" x from "doc.txt" if (x ", a full service")'
                       'with threshold 0.8',
                       ['Cafe Benz'])

    def test_ents_semantic_right_context(self):
        self.QueryTest('this is Cafe Benz, a full service cafe.',
                       'extract "Ents" x from "doc.txt" if (x ~ "full service")'
                       'with threshold 0.8',
                       ['Cafe Benz'])

    def test_nps_left_context(self):
        self.QueryTest("Should we go to Starbucks or Philz Coffee this time?",
                       'extract "NPs" x from "doc.txt" if ("go to" x)'
                       'with threshold 0.8',
                       ['Starbucks'])

    def test_nps_inside_context(self):
        self.QueryTest('Philz Coffee is an American coffeehouse chain.',
                       'extract "NPs" x from "doc.txt" if (str(x) contains "Coffee")'
                       'with threshold 0.7',
                       ['Philz Coffee'])

    def test_nps_inside_context_score(self):
        self.QueryTest('Philz Coffee is an American coffeehouse chain.',
                       'extract "NPs" x from "doc.txt" if '
                       '        (str(x) contains "Coffee" {0.8}) or'
                       '        (str(x) contains "coffeehouse" {0.9})'
                       'with threshold 0.7',
                       ['American coffeehouse chain', 'Philz Coffee'],
                       [0.9, 0.8])

    def test_nps_substring_context(self):
        self.QueryTest('Philz Coffee is an American coffeehouse chain.',
                       'extract "NPs" x from "doc.txt" if (str(x) mentions "hilz")'
                       'with threshold 0.7',
                       ['Philz Coffee'])

    def test_ngrams_right_near_context(self):
        self.QueryTest('This is Cafe Benz, a full service cafe.',
                       '''
                       extract "Ngrams(1,2)" x from "doc.txt" if
                       (x near "a full service" {0.5}) or
                       ("is" x {0.5})
                       with threshold 0.5
                       excluding (str(x) contains ",")
                       ''',
                       ['Cafe Benz', 'Cafe'])

    def test_excluding(self):
        self.QueryTest('Philz Coffee is a coffeehouse chain with many stores in California.',
                       'extract "NPs" x from "doc.txt" if '
                       '        (str(x) contains "Coffee" {0.8}) or'
                       '        (str(x) contains "coffeehouse" {0.9}) or'
                       '        ("in" x) '
                       'with threshold 0.7 '
                       'excluding (str(x) matches "[a-z ]+") or (str(x) in dict("Location"))',
                       ['Philz Coffee'],
                       [0.8])

    def test_multiple_condition_score(self):
        self.QueryTest('Philz Coffee is an American coffeehouse chain.',
                       'extract "NPs" x from "doc.txt" if'
                       '        (x near "coffeehouse" {0.8}) or'
                       '        (str(x) contains "Coffee" {0.95})'
                       'with threshold 0.8',
                       ['Philz Coffee'],
                       [1.0])

    def test_nps_precedes_context_score(self):
        self.QueryTest('In 2003, Phil converted his corner store into Philz Coffee.',
                       'extract "NPs" x from "doc.txt" if ("store" near x {0.8})'
                       'with threshold 0.3',
                       ['Philz Coffee'],
                       [0.4])

    def test_nps_follows_context_score(self):
        self.QueryTest('Philz Coffee has great coffee.',
                       'extract "NPs" x from "doc.txt" if (x near "coffee" {0.9})'
                       'with threshold 0.2',
                       ['Philz Coffee'],
                       [0.3])

    def test_exercise_score(self):
        self.QueryTest('I like figure skating on TV.',
                       'extract "NPs" x from "doc.txt" if (str(x) in dict("Exercise") {0.99})'
                       'with threshold 0.8',
                       ['figure skating'],
                       [0.99])

'''
    def test_ents_etype_context(self):
        self.QueryTest('Let me introduce Cafe Benz, a full service cafe '
                       'in the middle of the car dealership.'
                       'Sit on a soft leather couch in Cafe Benz.',
                       'extract x:Entity from "doc.txt" if ("introduce" x)'
                       'with threshold 0.8',
                       ['Cafe Benz'])
'''

if __name__ == '__main__':
    unittest.main()

    



'''
Copyright 2017 Recruit Institute of Technology

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import unittest
from koko.segmenter import Segmenter, SegmentedDocument

class SegmenterTestCase(unittest.TestCase):
                                   #            1         2
    def test_segment_query(self):  #  012345678901234567890123456
        segments = Segmenter.Segment('extract x from "doc.txt" if\n' +
                                          '  (str(x) contains "cafe")\n')

        expected_values = ['extract', 'x', 'from', '"', 'doc', '.', 'txt', '"', 'if',
                           '(', 'str', '(', 'x', ')', 'contains', '"', 'cafe', '"', ')']
        expected_rows = [1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2]
        expected_cols = [0, 8, 10, 15, 16, 19, 20, 23, 25, 2, 3, 6, 7, 8, 10, 19, 20, 24, 25]
        self.assertEqual(len(segments), len(expected_values))
        self.assertEqual(len(segments), len(expected_rows))
        self.assertEqual(len(segments), len(expected_cols))
        for i in range(len(segments)):
            self.assertEqual(segments[i].text, expected_values[i], 'token %d' % i)
            self.assertEqual(segments[i].row, expected_rows[i], 'token %d' % i)
            self.assertEqual(segments[i].col, expected_cols[i], 'token %d' % i)

    def test_segment_float(self):
        segments = Segmenter.Segment('Pi=3.14')
        expected_values = ['Pi', '=', '3.14']
        self.assertEqual(len(segments), len(expected_values))
        for i in range(len(segments)):
            self.assertEqual(segments[i].text, expected_values[i], 'token %d' % i)

    def test_segmented_document(self):
        #                        0 1    2  3    4       5  678  9  10  11    12   13    14 15  16
        doc = SegmentedDocument('I woke up this morning at 7:50 am and drove 3.14 miles to work.')
        self.assertEqual(len(doc), 17)
        self.assertEqual(doc[4].text, 'morning')
        self.assertEqual(doc[12].text, '3.14')
        time_span = doc[6:10]
        time_span.type = "time"
        self.assertEqual(time_span.text, '7:50 am')
        self.assertEqual(doc[6].text, '7')
        self.assertEqual(doc[7].text, ':')
        self.assertEqual(doc[8].text, '50')
        
            
if __name__ == '__main__':
    unittest.main()




'''
Copyright 2017 Recruit Institute of Technology

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import os

class DictionaryEntityClassifier:
    def __init__(self, filename):
        self.dictionary = {}
        self.load_dictionary(filename)

    def load_dictionary(self, filename):
        if not filename:
            return
        if not os.path.exists(filename):
            print('File not found: %s' % filename)
            return

        file = open(filename)
        for line in file.readlines():
            self.dictionary[line.rstrip()] = 1
        file.close()

    def is_entity(self, name):
        if name in self.dictionary:
            return self.dictionary[name]
        else:
            return 0.0



'''
Copyright 2017 Recruit Institute of Technology

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from .matcher import Matcher, Ngram
from .parser import Parser
from .scorer import Scorer

import sys

class Entity:

    def __init__(self, span):
        self.span = span
        self.mentions = []
        self.score = 0.0
        self.scores = []
        self.name = span.text.replace('\n', ' ')

    def strip(self):
        del self.scores
        for mention in self.mentions:
            mention.strip()

class Mention:

    def __init__(self, span, sentence_index=-1):
        self.span = span
        self.sentence_index = sentence_index
        self.score = 0.0
        self.scores = []
        self.debug = ''

    def strip(self):
        del self.span
        del self.sentence_index
        del self.scores
        del self.debug


class EntityExtractor:

    def __init__(self, doc, testing=False):
        self.doc = doc
        assert self.doc.is_parsed
        self.error_msg = None
        self.testing = testing
        self.matcher = Matcher(self.doc)
        #print("Collecting set of document words")
        self.doc_words = set()
        for i in range(len(doc)):
            self.doc_words.add(doc[i].text.lower())

    def TopEntities(self, query, target_lang='en', embedding_path=None):
        #print("Parse the query")
        parser = Parser(query, target_lang, self.doc_words, self.testing, embedding_path)
        if not parser.is_parsed:
            self.error_msg = parser.error_msg
            return []
        return self.TopEntitiesForParsedQuery(parser)
        
    def TopEntitiesForParsedQuery(self, parser):
        self.query_debug = parser.toString()
        spans = self.GetSpans(parser.etype)
        #print("GetMentionsFromSpans")
        #all_mentions = self.GetMentionsFromSpans(spans)
        all_mentions = self.JoinSpansAndSentences(spans)
        return self.TopEntitiesFromMentions(parser, all_mentions)
        
    def TopEntitiesFromMentions(self, parser, mentions):
        scorer = Scorer(self.doc, self.matcher,
                        parser.predicates,
                        parser.excluding_predicates,
                        parser.sentence_decomposer_server_url)
        #print("ScoreMentions")
        scorer.ScoreMentions(mentions)
        #print("FilterMentions")
        filtered_mentions = self.FilterMentions(mentions)
        #print("ClusterMentions")
        entities = self.ClusterMentions(filtered_mentions)
        #print("ScoreEntities")
        scorer.ScoreEntities(entities)
        filtered_entities = self.FilterEntities(entities, parser.threshold)
        #self.StripEntities(filtered_entities)
        return sorted(filtered_entities, key=lambda x: x.score, reverse=True)


    def GetSpans(self, etype):
        # Collect all spans of the given type
        spans = []
        if etype == "Ents":
            spans = self.doc.ents
        elif etype == "NPs":
            spans = self.doc.noun_chunks
        elif etype[:6] == "Ngrams":
            bounds = [int(x) for x in etype[7:-1].split(',')]
            for length in range(bounds[0], bounds[1] + 1):
                for i in range(len(self.doc)-length):
                    spans.append(self.doc[i:i+length])
        else:
            etype = etype.lower()
            for i in range(len(self.doc.ents)):
                if etype == "entity" or self.doc.ents[i].syntax_type == etype:
                    spans.append(self.doc.ents[i])
        return sorted(spans, key=lambda x: x.start, reverse=False)

    def GetMentionsFromSpans(self, spans):
        mentions = []
        for span in spans:
            si = self.GetSpanSentenceNumber(span)
            mention = Mention(span)
            mention.sentence_index = si
            mentions.append(mention)
        return mentions

    def JoinSpansAndSentences(self, spans):
        mentions = []
        sents = [sent for sent in self.doc.sents]
        si = 0
        i = 0
        while i < len(spans):
            #print(i, si)
            sent = sents[si]
            span = spans[i]
            #print(span.start, span.end, sent.start, sent.end)
            if sent.end <= span.start:
                si += 1
                continue
            if span.end <= sent.end:
                mentions.append(Mention(span, si))
            i += 1
        return mentions
    
    def FindAllMentions(self, spans):
        print("GenerateAllMentions")
        mentions = []
        print("Deduplicate ngrams")
        ngrams = set(Ngram(self.doc, span) for span in spans)
        ## This is very slow for large documents.
        for ngram in ngrams:
            print(' '.join(ngram.tokens()))
            n = len(ngram)
            si = -1
            for sent in self.doc.sents:
                si += 1
                for i in range(sent.start, sent.end - n):
                    if self.matcher.MatchesNgram(i, ngram.tokens()):
                        mentions.append(Mention(self.doc[i: i + n], si))
        return mentions

    def FilterMentions(self, mentions):
        return [mention for mention in mentions if mention.score > 0]

    def FilterEntities(self, entities, threshold):
        return [e for e in entities if e.score >= threshold]

    def StripEntities(self, entities):
        for e in entities:
            e.strip()

    def ClusterMentions(self, mentions):
        ent_dict = {}
        for mention in mentions:
            name = mention.span.text.lower()
            if name not in ent_dict:
                ent_dict[name] = Entity(mention.span)
            ent_dict[name].mentions.append(mention)
        return ent_dict.values()

    def GetSentence(self, mention):
        for sent in self.doc.sents:
            if mention.span.start >= sent.start and mention.span.end <= sent.end:
                return sent.text
        return "None"

    def GetSpanSentenceNumber(self, span):
        sentence_index = -1
        for sent in self.doc.sents:
            sentence_index += 1
            if span.start >= sent.start and span.end <= sent.end:
                return sentence_index
        return -1
    
    def GetSentenceNumber(self, mention):
        sentence_index = -1
        for sent in self.doc.sents:
            sentence_index += 1
            if mention.span.start >= sent.start and mention.span.end <= sent.end:
                return sentence_index
        return -1



'''
Copyright 2017 Recruit Institute of Technology

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

# Unit tests for EntityExtractor

# Warning: the extracted entities and noun chunks are fake, using simple heuristics for extraction.

import unittest
from koko.test_document import TestDocument  

class TestDocumentTestCase(unittest.TestCase):
    def setUp(self):
        #           0 1    2     3      4    5    6        7 8   9   10   11   12    13     14  15 16
        document = 'I like Philz Coffee more than Starbucks, and way more than Peets Coffee and Tea.'
        self.doc = TestDocument(document)
    
    def test_len(self):
        self.assertEqual(len(self.doc), 17)

    def test_token(self):
        token = self.doc[3]
        self.assertEqual(token.start, 13)
        self.assertEqual(token.end, 19)
        self.assertEqual(token.text, 'Coffee')

    def test_span(self):
        span = self.doc[2:4]
        self.assertEqual(span.start, 2)
        self.assertEqual(span.end, 4)
        self.assertEqual(span.text, 'Philz Coffee')

    def test_span_text(self):
        span = self.doc[6:9]
        self.assertEqual(span.start, 6)
        self.assertEqual(span.end, 9)
        self.assertEqual(span.text, 'Starbucks, and')

    def test_ents(self):
        entities = [e.text for e in self.doc.ents]
        self.assertEqual(entities, ['I', 'Philz Coffee', 'Starbucks', 'Peets Coffee', 'Tea'])

    def test_noun_chunks(self):
        nps = [e.text for e in self.doc.noun_chunks]
        self.assertEqual(nps, ['Philz Coffee', 'Starbucks', 'Peets Coffee'])

    def test_sents(self):
        sents = self.doc.sents
        self.assertEqual(len(sents), 1)
        sent = sents[0]
        self.assertEqual(sent.start, 0)
        self.assertEqual(sent.end, len(self.doc))
        self.assertEqual(sent.text,
                         'I like Philz Coffee more than Starbucks, '
                         'and way more than Peets Coffee and Tea.')

if __name__ == '__main__':
    unittest.main()



"""
Copyright 2017 Recruit Institute of Technology

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import sys
import logging
from .query_processor import QueryProcessor


def run(query_file, doc_parser="koko", output_format="text", log_level="info", target_lang='en', embedding_file="../input/embedding.txt"):
    # Set up logging
    logging_level_dict = {'info': logging.INFO,
                          'warning': logging.WARNING,
                          'error': logging.ERROR}
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    ch = logging.StreamHandler(sys.stdout)
    ch.setLevel(logging_level_dict[log_level])
    formatter = logging.Formatter('%(levelname)s %(asctime)s - %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    with open(query_file, 'r') as qfile:
        query = qfile.read()

    # Process the KOKO query

    processor = QueryProcessor(doc_parser, target_lang)
    response = processor.ProcessQuery(query, embedding_path=embedding_file)

    # Print the results
    if output_format == 'text':
        print("\nResults:\n")
        #print("%s %s" % ("{:<50}".format("Entity name"), "Entity score"))
        print("%s %s %s" % ("{:<30}".format("Entity name"), "{:<20}".format("Entity count"), "Entity score"))        
        #print("===============================================================")
        print("="*70)
        for entity in response.entities:
            #print("%s %f" % ("{:<50}".format(entity.name), entity.score))
            print("%s %s %f" % ("{:<30}".format(entity.name), "{:<20}".format(str(len(entity.mentions))), entity.score))
    else:
        import json
        import jsonpickle
        pickled = jsonpickle.encode(response, unpicklable=False)
        json_result = json.loads(pickled)
        print(json.dumps(json_result, sort_keys=False, indent=2))
        with open('json_result.txt', 'w') as ofile:
            json.dump(json_result, ofile)



'''
Copyright 2017 Recruit Institute of Technology

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from .matcher import Matcher, Ngram
from .sentence_decomposer import SentenceDecomposer

class SemanticScorer(object):

    def __init__(self, doc, sentence_decomposer_server_url = ''):
        self.doc = doc
        self.sentence_wordsets = []
        self.sentences = []
        self.entailed_sentences = []
        self.sentence_decomposer_server_url = sentence_decomposer_server_url
        for sent in doc.sents:
            wordset = set()
            for word in sent:
                wordset.add(word.text.lower())
            self.sentence_wordsets.append(wordset)
            self.sentences.append(sent)
            self.entailed_sentences.append([])

    def PruneExpandedQueries(self, sentence_index, expanded_queries):
        valid_expanded_queries = []
        for (expanded_tokens, query_score) in expanded_queries:
            valid = True
            for token in expanded_tokens:
                if token not in self.sentence_wordsets[sentence_index]:
                    valid = False
                    break
            if valid:
                valid_expanded_queries.append((expanded_tokens, query_score))
        return valid_expanded_queries

    def ComputeSemanticScore(self, mention, predicate):
        ngram = Ngram(self.doc, mention.span).tokens()
        maximum_score = 0.0
        sentence_index = mention.sentence_index
        valid_expanded_queries = self.PruneExpandedQueries(sentence_index,
                                                           predicate.expanded_queries)
        if not valid_expanded_queries:
            return 0.0
        
        if predicate.type == 'inside':
            matcher = Matcher(self.doc)
            for (expanded_tokens, query_score) in valid_expanded_queries:
                if query_score <= maximum_score:
                    continue
                for pos in range(mention.span.start, \
                                 mention.span.end - len(expanded_tokens) + 1):
                    if matcher.MatchesNgram(pos, expanded_tokens):
                        maximum_score = query_score
            return maximum_score
        elif predicate.type == 'substring':
            for (expanded_tokens, query_score) in valid_expanded_queries:
                if query_score <= maximum_score:
                    continue
                if mention.span.text.lower().find(' '.join(expanded_tokens)) >= 0:
                    maximum_score = query_score
            return maximum_score
        

        entailed_sentences = self.entailed_sentences[sentence_index]  # look up in cache
        if not entailed_sentences:
            # lazy decomposition
            sentence = self.sentences[sentence_index]
            # print('Requesting entailed sentences for: %s' % sentence)
            
            decomposer = SentenceDecomposer(self.sentence_decomposer_server_url)
            entailed_sentences = decomposer.decompose(sentence)
            self.entailed_sentences[sentence_index] = entailed_sentences  # cache the result

        for entailed_sentence_list in entailed_sentences:
            for entailed_sentence in entailed_sentence_list:
                if entailed_sentence.score <= maximum_score:
                    continue
                matcher = Matcher(entailed_sentence.doc)
                start, end = matcher.FindNgram(ngram)
                if start == end:
                    continue
                for (expanded_tokens, query_score) in valid_expanded_queries:
                    score = query_score * entailed_sentence.score
                    if score <= maximum_score:
                        continue
                    sent = entailed_sentence.doc
                    sequence = sent[0:start]
                    if predicate.type == 'right':
                        if end == len(sent):
                            continue
                        sequence = sent[end:len(sent)]
                    #if self.SubsequenceMatch(sequence, expanded_tokens):
                    if self.SubsequenceExactMatch(sequence, expanded_tokens, predicate.type):
                        maximum_score = score
                        mention.debug = '%s : %s' % (' '.join(expanded_tokens),
                                                     entailed_sentence.doc.text)
        return maximum_score

    def SubsequenceMatch(self, sequence, pattern):
        start = 0
        end = len(sequence)
        for token in pattern:
            found = False
            for pos in range(start, end):
                if sequence[pos].text.lower() == token.lower():
                    found = True
                    start = pos + 1
                    break
            if not found:
                return False
        return True

    def SubsequenceExactMatch(self, sequence, pattern, pred_type):
        "Determine if 'pattern' is a exact subsequence of 'sequence'"
        start = len(sequence) - len(pattern)
        if start < 0:
            return False

        if pred_type == 'right':
            start = 0

        for offset in range(len(pattern)):
            sent_idx = start + offset
            if sequence[sent_idx].text.lower() != pattern[offset].lower():
                return False
        return True



'''
Copyright 2017 Recruit Institute of Technology

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

# coding=utf-8

import string


class Segment:

    def __init__(self, text, row, col, start, end, append_to_previous):
        self.row = row
        self.col = col
        self.start = start
        self.end = end
        self.text = text[start:end]
        self.append_to_previous = append_to_previous

    def toString(self):
        return 'Line %d col %d start %d end %d: \'%s\'' % (self.row, self.col,
                                                           self.start, self.end, self.text)


class Segmenter:

    def __init__(self):
        pass

    @staticmethod
    def Segment(text):
        segments = []
        in_segment = False
        lines = text.splitlines(True)
        row = 0
        punctuation = set(string.punctuation) | \
                      set(['“', '”', '‘', '’'])
        number_punctuation = set(['.', ','])
        line_start = 0
        n = 0
        for i in range(len(lines)):
            row += 1
            line = lines[i]
            col = 0  # holds start of segment
            line_start += n
            n = len(line)

            for j in range(n):
                c = line[j]
                if c.isspace() and in_segment:
                    append_to_previous = segments and segments[-1].end == line_start + col
                    segments.append(
                        Segment(text, row, col, line_start + col, line_start + j,
                                append_to_previous))
                    append_to_previous = False
                    in_segment = False
                elif (j > 0 and line[j - 1].isdigit() and
                      c in number_punctuation and
                      j + 1 < n and line[j + 1].isdigit()):
                    pass
                elif c in punctuation:
                    if in_segment:
                        # terminate current segment
                        in_segment = False
                        append_to_previous = segments and segments[-1].end == line_start + col
                        segments.append(
                            Segment(text, row, col, line_start + col, line_start + j,
                                    append_to_previous))
                    # create new segment with the punctuation character
                    append_to_previous = segments and segments[-1].end == line_start + j
                    segments.append(
                        Segment(text, row, j, line_start + j, line_start + j + 1,
                                    append_to_previous))
                    in_segment = False
                elif not c.isspace() and not in_segment:
                    # mark start of new segment
                    # print('column %d char %s start segment' % (j, c))
                    col = j
                    in_segment = True
        if in_segment:
            append_to_previous = segments and segments[-1].end == line_start + col
            segments.append(
                Segment(text, row, col, line_start + col, line_start + j + 1,
                        append_to_previous))
        return segments


class SegmentedSpan:

    def __init__(self, doc, start, end, syntax_type=''):
        assert(start < len(doc) and start >= 0 and
               end <= len(doc) and end >= 0), \
               "Span %d:%d is out of the document range 0:%d." % (start, end, len(doc))
        self.doc = doc
        self.start = start
        self.end = end
        first_segment = doc.segments[start]
        last_segment = doc.segments[end - 1]
        self.row = first_segment.row
        self.col = first_segment.col
        self.start_char = first_segment.start
        self.end_char = last_segment.end
        self.segments = doc.segments[start:end]
        self.text = doc.text[first_segment.start:last_segment.end]
        self.type = "text"
        self.syntax_type = syntax_type

    def __len__(self):
        return self.end - self.start

    def __getitem__(self, x):
        if isinstance(x, slice):
            return SegmentedSpan(self.doc, self.start + x.start, self.start + x.stop)
        return self.segments[x]


class SegmentedDocument:

    def __init__(self, text):
        self.text = text
        self.segments = Segmenter.Segment(self.text)

    def __len__(self):
        return len(self.segments)

    def __getitem__(self, x):
        if isinstance(x, slice):
            return SegmentedSpan(self, x.start, x.stop)
        return SegmentedSpan(self, x, x + 1)[0]



'''
Copyright 2017 Recruit Institute of Technology

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from .entity_extractor import EntityExtractor
from .parser import Parser
from .koko_document import KokoDocument
import sys

def run(query_file):
    with open(query_file, 'r') as qfile:
        query = qfile.read()
    print("Query:\n%s" % query)

    parser = Parser(query)
    if not parser.is_parsed:
        print('Syntax error: ', parser.error_msg)
        return

    print("Rewritten query:\n", parser.toString())

    with open(parser.document_name, 'r') as myfile:
        document = myfile.read()

    extractor = EntityExtractor(KokoDocument(document), True)
    entities = extractor.TopEntities(query)
    print("\nResults:\n")
    print("%s %s" % ("{:<50}".format("Entity name"), "Entity score"))
    print("==================================================================")
    for entity in entities:
        print("%s %f" % ("{:<50}".format(entity.span.text), entity.score))


# main

def main():
    if len(sys.argv) < 2:
        print("Usage: python koko.py *.koko")
        exit()

    query_files = sys.argv[1:]

    for query_file in query_files:
        run(query_file)

if __name__ == '__main__':
    main()
